@@ -2,16 +2,17 @@
 #
 # This file is part of gunicorn released under the MIT license.
 # See the NOTICE for more information.
+from __future__ import print_function
 
 import os
 import sys
 import traceback
 
+from gunicorn._compat import execfile_
 from gunicorn import util
 from gunicorn.arbiter import Arbiter
 from gunicorn.config import Config, get_default_config_file
 from gunicorn import debug
-from gunicorn.six import execfile_
 
 class BaseApplication(object):
     """
@@ -34,7 +35,7 @@ class BaseApplication(object):
             self.load_default_config()
             self.load_config()
         except Exception as e:
-            sys.stderr.write("\nError: %s\n" % str(e))
+            print("\nError: %s" % str(e), file=sys.stderr)
             sys.stderr.flush()
             sys.exit(1)
 
@@ -70,16 +71,27 @@ class BaseApplication(object):
         try:
             Arbiter(self).run()
         except RuntimeError as e:
-            sys.stderr.write("\nError: %s\n\n" % e)
+            print("\nError: %s\n" % e, file=sys.stderr)
             sys.stderr.flush()
             sys.exit(1)
 
+
 class Application(BaseApplication):
-    def load_config_from_file(self, filename):
-        """
-        Loads the configuration file: the file is a python file, otherwise raise an RuntimeError
-        Exception or stop the process if the configuration file contains a syntax error.
-        """
+
+    # 'init' and 'load' methods are implemented by WSGIApplication.
+    # pylint: disable=abstract-method
+
+    def chdir(self):
+        # chdir to the configured path before loading,
+        # default is the current dir
+        os.chdir(self.cfg.chdir)
+
+        # add the path to sys.path
+        if self.cfg.chdir not in sys.path:
+            sys.path.insert(0, self.cfg.chdir)
+
+    def get_config_from_filename(self, filename):
+
         if not os.path.exists(filename):
             raise RuntimeError("%r doesn't exist" % filename)
 
@@ -93,10 +105,32 @@ class Application(BaseApplication):
         try:
             execfile_(filename, cfg, cfg)
         except Exception:
-            print("Failed to read config file: %s" % filename)
+            print("Failed to read config file: %s" % filename, file=sys.stderr)
             traceback.print_exc()
+            sys.stderr.flush()
             sys.exit(1)
 
+        return cfg
+
+    def get_config_from_module_name(self, module_name):
+        return vars(util.import_module(module_name))
+
+    def load_config_from_module_name_or_filename(self, location):
+        """
+        Loads the configuration file: the file is a python file, otherwise raise an RuntimeError
+        Exception or stop the process if the configuration file contains a syntax error.
+        """
+
+        if location.startswith("python:"):
+            module_name = location[len("python:"):]
+            cfg = self.get_config_from_module_name(module_name)
+        else:
+            if location.startswith("file:"):
+                filename = location[len("file:"):]
+            else:
+                filename = location
+            cfg = self.get_config_from_filename(filename)
+
         for k, v in cfg.items():
             # Ignore unknown names
             if k not in self.cfg.settings:
@@ -104,11 +138,15 @@ class Application(BaseApplication):
             try:
                 self.cfg.set(k.lower(), v)
             except:
-                sys.stderr.write("Invalid value for %s: %s\n\n" % (k, v))
+                print("Invalid value for %s: %s\n" % (k, v), file=sys.stderr)
+                sys.stderr.flush()
                 raise
 
         return cfg
 
+    def load_config_from_file(self, filename):
+        return self.load_config_from_module_name_or_filename(location=filename)
+
     def load_config(self):
         # parse console args
         parser = self.cfg.parser()
@@ -117,33 +155,52 @@ class Application(BaseApplication):
         # optional settings from apps
         cfg = self.init(parser, args, args.args)
 
+        # set up import paths and follow symlinks
+        self.chdir()
+
         # Load up the any app specific configuration
-        if cfg and cfg is not None:
+        if cfg:
             for k, v in cfg.items():
                 self.cfg.set(k.lower(), v)
 
+        env_args = parser.parse_args(self.cfg.get_cmd_args_from_env())
+
         if args.config:
             self.load_config_from_file(args.config)
+        elif env_args.config:
+            self.load_config_from_file(env_args.config)
         else:
             default_config = get_default_config_file()
             if default_config is not None:
                 self.load_config_from_file(default_config)
 
-        # Lastly, update the configuration with any command line
-        # settings.
-        for k, v in args.__dict__.items():
+        # Load up environment configuration
+        for k, v in vars(env_args).items():
             if v is None:
                 continue
             if k == "args":
                 continue
             self.cfg.set(k.lower(), v)
 
+        # Lastly, update the configuration with any command line settings.
+        for k, v in vars(args).items():
+            if v is None:
+                continue
+            if k == "args":
+                continue
+            self.cfg.set(k.lower(), v)
+
+        # current directory might be changed by the config now
+        # set up import paths and follow symlinks
+        self.chdir()
+
     def run(self):
         if self.cfg.check_config:
             try:
                 self.load()
             except:
-                sys.stderr.write("\nError while loading the application:\n\n")
+                msg = "\nError while loading the application:\n"
+                print(msg, file=sys.stderr)
                 traceback.print_exc()
                 sys.stderr.flush()
                 sys.exit(1)
@@ -156,7 +213,7 @@ class Application(BaseApplication):
             util.daemonize(self.cfg.enable_stdio_inheritance)
 
         # set python paths
-        if self.cfg.pythonpath and self.cfg.pythonpath is not None:
+        if self.cfg.pythonpath:
             paths = self.cfg.pythonpath.split(",")
             for path in paths:
                 pythonpath = os.path.abspath(path)
